package ru.dbotthepony.mc.otm.core.util

import it.unimi.dsi.fastutil.HashCommon
import net.minecraft.util.Mth
import net.minecraft.util.RandomSource
import net.minecraft.world.level.levelgen.MarsagliaPolarGaussian
import net.minecraft.world.level.levelgen.PositionalRandomFactory
import net.minecraft.world.level.levelgen.RandomSupport
import java.util.random.RandomGenerator

/**
 * Excellent number generator with guaranteed period of 2^255
 */
class Xoshiro256SSRandom private constructor(
	private var s0: Long,
	private var s1: Long,
	private var s2: Long,
	private var s3: Long,
	marker: Nothing?
) : RandomGenerator, RandomSource {
	private val gaussian = MarsagliaPolarGaussian(this)

	init {
		if (s0 or s1 or s2 or s3 == 0L) {
			s0 = 0x73CF3D83FFF44FF3L
			s1 = 0x6412312B70F3CD37L
			s2 = -0X6BB4C4E1327BFDCFL
			s3 = -0X4BE0F5BB5F3F5240L
		}
	}

	constructor(s0: Long, s1: Long, s2: Long, s3: Long) : this(s0, s1, s2, s3, null) {
		// discard some values so generator can get going if provided seed was "weak"
		for (i in 0 until 32)
			nextLong()
	}

	constructor(seed: Long) : this(1L, 2L, 3L, 4L, null) {
		setSeed(seed)
	}

	constructor() : this(RandomSupport.generateUniqueSeed(), RandomSupport.generateUniqueSeed(), RandomSupport.generateUniqueSeed(), RandomSupport.generateUniqueSeed(), null)

	override fun setSeed(seed: Long) {
		val rng = LCG64Random(seed)
		s0 = rng.nextLong()
		s1 = rng.nextLong()
		s2 = rng.nextLong()
		s3 = rng.nextLong()
		gaussian.reset()
	}

	override fun nextInt(): Int {
		// sample upper bits
		return nextLong().ushr(32).toInt()
	}

	override fun nextLong(): Long {
		val result = (s1 * 5).rotateLeft(7) * 9
		val t = s1.shl(17)
		s2 = s2.xor(s0)
		s3 = s3.xor(s1)
		s1 = s1.xor(s2)
		s0 = s0.xor(s3)
		s2 = s2.xor(t)
		s3 = s3.rotateLeft(45)
		return result
	}

	override fun nextInt(bound: Int): Int {
		return super<RandomGenerator>.nextInt(bound)
	}

	override fun nextInt(origin: Int, bound: Int): Int {
		return super<RandomGenerator>.nextInt(origin, bound)
	}

	override fun nextBoolean(): Boolean {
		return super.nextBoolean()
	}

	override fun nextFloat(): Float {
		return super.nextFloat()
	}

	override fun nextDouble(): Double {
		return super.nextDouble()
	}

	override fun nextGaussian(): Double {
		return gaussian.nextGaussian()
	}

	override fun fork(): RandomSource {
		return Xoshiro256SSRandom(nextLong(), nextLong(), nextLong(), nextLong(), null)
	}

	override fun forkPositional(): PositionalRandomFactory {
		return Positional(nextLong(), nextLong(), nextLong(), nextLong())
	}

	class Positional(
		val s0: Long,
		val s1: Long,
		val s2: Long,
		val s3: Long,
	) : PositionalRandomFactory {
		override fun at(x: Int, y: Int, z: Int): RandomSource {
			val rng = LCG64Random(Mth.getSeed(x, y, z))

			return Xoshiro256SSRandom(
				s0.rotateLeft(11).xor(rng.nextLong()),
				s1.rotateLeft(22).xor(rng.nextLong()),
				s2.rotateLeft(33).xor(rng.nextLong()),
				s3.rotateLeft(44).xor(rng.nextLong()),
			)
		}

		override fun fromHashOf(name: String): RandomSource {
			return Xoshiro256SSRandom(s0, HashCommon.murmurHash3(name.hashCode().toLong()).xor(s1), s2, s3)
		}

		override fun fromSeed(seed: Long): RandomSource {
			return Xoshiro256SSRandom(seed)
		}

		override fun parityConfigString(builder: StringBuilder) {
			throw UnsupportedOperationException()
		}
	}

	companion object {
		@JvmStatic
		fun raw(s0: Long, s1: Long, s2: Long, s3: Long): Xoshiro256SSRandom {
			return Xoshiro256SSRandom(s0, s1, s2, s3, null)
		}
	}
}
